use derive_more::derive;

use super::range_mapping::RangeMapping;

#[derive(Debug)]
pub struct MappingLayer {
    label: String,
    ranges: Vec<RangeMapping>,
}

impl MappingLayer {
    pub fn new(label: impl Into<String>, mut ranges: Vec<RangeMapping>) -> Self {
        ranges.sort_by_key(|element| element.source());
        let ranges = Self::fill_in_the_gaps(ranges);
        Self {
            label: label.into(),
            ranges,
        }
    }

    /// # Expected
    /// It is assumed that ranges is sorted by the field `source`
    pub fn fill_in_the_gaps(ranges: Vec<RangeMapping>) -> Vec<RangeMapping> {
        let mut current_source_start = 0;
        let mut without_gaps: Vec<RangeMapping> = Vec::new();
        for next_range in ranges {
            let next_source = next_range.source();

            assert!(current_source_start <= next_source);
            if current_source_start == next_source {
                current_source_start = next_range.inclusive_end_source().saturating_add(1);
                without_gaps.push(next_range);
            } else {
                let new_range = next_source;
                without_gaps.push(RangeMapping::just_with_source(
                    current_source_start,
                    new_range,
                ));
                current_source_start = next_range.inclusive_end_source() + 1;
                without_gaps.push(next_range);
            }
        }

        without_gaps.push(RangeMapping::source_to_max(current_source_start));

        without_gaps
    }

    pub fn ranges(&self) -> &[RangeMapping] {
        &self.ranges
    }
}

#[cfg(test)]
mod testing {
    use crate::solutions::day5::{mapping_layer::MappingLayer, range_mapping::RangeMapping};

    #[test]
    fn should_fill_in_the_gaps() {
        let input = vec![
            RangeMapping::new(45, 81, 19),
            RangeMapping::new(64, 68, 13),
            RangeMapping::new(77, 45, 23),
            RangeMapping::new(110, 55, 20),
        ];
        let actual = MappingLayer::fill_in_the_gaps(input);
        insta::assert_debug_snapshot!(actual);
    }
}
