use std::{str::FromStr, u32};

use thiserror::Error;

use super::UnsignedNumber;

#[derive(Debug, PartialEq, Eq, Clone, Copy)]
pub struct RangeMapping {
    source: UnsignedNumber,
    target: UnsignedNumber,
    range: UnsignedNumber,
}

#[derive(Debug, Error)]
pub enum ParseErrorRangeMapping {
    #[error("Needs to have at least 3 numbers")]
    LessThan3Numbers,
    #[error("Every item must be a valid unsigned number")]
    InvalidFormatForNumbers,
}

impl FromStr for RangeMapping {
    type Err = ParseErrorRangeMapping;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let mut numbers_as_3 = s.split(" ").take(3).map(|unparsed| {
            unparsed
                .parse::<UnsignedNumber>()
                .map_err(|_| ParseErrorRangeMapping::InvalidFormatForNumbers)
        });
        let (target, source, range) = (
            numbers_as_3
                .next()
                .ok_or(ParseErrorRangeMapping::LessThan3Numbers)??,
            numbers_as_3
                .next()
                .ok_or(ParseErrorRangeMapping::LessThan3Numbers)??,
            numbers_as_3
                .next()
                .ok_or(ParseErrorRangeMapping::LessThan3Numbers)??,
        );

        Ok(Self {
            target,
            source,
            range,
        })
    }
}

impl RangeMapping {
    pub fn new(source: UnsignedNumber, target: UnsignedNumber, range: UnsignedNumber) -> Self {
        Self {
            source,
            target,
            range,
        }
    }

    pub fn just_with_source(source: UnsignedNumber, range: UnsignedNumber) -> Self {
        Self {
            source,
            target: source,
            range,
        }
    }

    pub fn source_to_max(source: UnsignedNumber) -> Self {
        const ENSURES_SOURCE_END_IS_MAX: UnsignedNumber = 1;
        let range = if source == 0 {
            UnsignedNumber::MAX
        } else {
            UnsignedNumber::MAX - (source - ENSURES_SOURCE_END_IS_MAX)
        };

        Self::just_with_source(source, range)
    }

    pub fn source(&self) -> UnsignedNumber {
        self.source
    }

    pub fn inclusive_end_source(&self) -> UnsignedNumber {
        if self.range() == UnsignedNumber::MAX {
            UnsignedNumber::MAX
        } else {
            self.source() + (self.range() - 1)
        }
    }

    pub fn map_position(&self, position: UnsignedNumber) -> UnsignedNumber {
        let inclusive_end_source = self.inclusive_end_source();
        if self.source() > position || position > inclusive_end_source {
            panic!(
                "Given position ({}) is outside of this range between ({}) and ({})",
                position,
                self.source(),
                inclusive_end_source
            );
        }
        let difference = position - self.source();
        self.target() + difference
    }

    pub fn target(&self) -> UnsignedNumber {
        self.target
    }

    pub fn range(&self) -> UnsignedNumber {
        self.range
    }
}

#[cfg(test)]
mod testing {
    use std::u32;

    use super::*;

    #[test]
    fn parse_from_line() {
        const INPUT: &str = "50 98 2";
        let expected = RangeMapping {
            source: 98,
            target: 50,
            range: 2,
        };
        let actual: RangeMapping = INPUT.parse().unwrap();
        assert_eq!(expected, actual);
    }

    #[test]
    fn calc_inclusive_end_source() {
        fn assert_case(given: RangeMapping, expected: UnsignedNumber) {
            let actual = given.inclusive_end_source();
            assert_eq!(expected, actual, "Given: {:#?}", given);
        }
        assert_case(RangeMapping::new(98, 50, 2), 99);
        assert_case(RangeMapping::new(0, 50, 1), 0);
        assert_case(RangeMapping::new(10, 50, 5), 14);
    }

    #[test]
    fn create_from_source_to_max_integer_max_correctly() {
        fn assert_case(input: RangeMapping) {
            const EXPECTED: UnsignedNumber = UnsignedNumber::MAX;
            let actual = input.inclusive_end_source();
            assert_eq!(EXPECTED, actual, "Given: {:#?}", input);
        }
        assert_case(RangeMapping::source_to_max(UnsignedNumber::MAX));
        assert_case(RangeMapping::source_to_max(UnsignedNumber::MAX - 1));
        assert_case(RangeMapping::source_to_max(100));
        assert_case(RangeMapping::source_to_max(2));
        assert_case(RangeMapping::source_to_max(0));
    }
}
